from mokr.constants.info import (  # noqa
    CHROME_VERSION,
    DEFAULT_CHROME_VERSION,
    FIREFOX_BUILD,
    INSTALL_DIR_NAME,
    INSTALL_PATH,
    MOKR_VERSION,
)
from mokr.constants.keys import KEY_DEFINITIONS  # noqa
from mokr.constants.events import (  # noqa
    BROWSER_CLOSE,
    BROWSER_GET_VERSION,
    CLOSE,
    DISCONNECTED,
    DOM_BOX_MODEL,
    DOM_CONTENT_QUADS,
    DOM_DESCRIBE_NODE,
    DOM_INPUT_FILES,
    DOM_LOADED,
    EMULATION_DISABLE_SCRIPT_EXECUTION,
    EMULATION_ENABLE_TOUCH,
    EMULATION_OVERRIDE_BACKGROUND,
    EMULATION_OVERRIDE_METRICS,
    EMULATION_SET_EMULATED_MEDIA,
    ERROR,
    FETCH_AUTH_REQD,
    FETCH_CONTINUE,
    FETCH_CONTINUE_AUTH,
    FETCH_DISABLED,
    FETCH_DOMAIN_REQUEST_SENT,
    FETCH_ENABLED,
    FETCH_FAIL,
    FETCH_FULFILL,
    FETCH_REQUEST_PAUSED,
    FRAME_ATTACHED,
    FRAME_DETACHED,
    FRAME_NAVIGATED,
    FRAME_NAVIGATED_IN_DOC,
    INPUT_INSERT_TEXT,
    INPUT_KEY,
    INPUT_MOUSE,
    INPUT_TOUCH,
    INSPECTOR_TARGET_CRASHED,
    LIFECYCLE_EVENT,
    LOG_ENABLE,
    LOG_ENTRY_ADDED,
    METRICS,
    NETWORK_CACHE_DISABLE,
    NETWORK_DELETE_COOKIES,
    NETWORK_EMULATE_NETWORK_CONDITIONS,
    NETWORK_ENABLE,
    NETWORK_EXTRA_HEADERS,
    NETWORK_GET_ALL_COOKIES,
    NETWORK_GET_COOKIES,
    NETWORK_GET_RESPONSE_BODY,
    NETWORK_LOADING_FAILED,
    NETWORK_LOADING_FINISHED,
    NETWORK_MGR_REQUEST,
    NETWORK_MGR_REQUEST_FAILED,
    NETWORK_MGR_REQUEST_FINISHED,
    NETWORK_MGR_REQUEST_FROM_CACHE,
    NETWORK_MGR_RESPONSE,
    NETWORK_REQUEST,
    NETWORK_REQUEST_FAILED,
    NETWORK_REQUEST_FINISHED,
    NETWORK_REQUEST_SERVED_FROM_CACHE,
    NETWORK_REQUEST_WILL_BE_SENT,
    NETWORK_RESPONSE,
    NETWORK_RESPONSE_RECVD,
    NETWORK_RESPONSE_RECVD_EXTRA,
    NETWORK_SET_COOKIES,
    NETWORK_USER_AGENT_OVERRIDE,
    PAGE_ADD_SCRIPT_TO_EVAL,
    PAGE_BRING_TO_FRONT,
    PAGE_CLOSE,
    PAGE_CONSOLE,
    PAGE_DIALOG,
    PAGE_DOM_LOADED,
    PAGE_ENABLE,
    PAGE_ENABLE_LIFECYCLE_EVENTS,
    PAGE_ERROR,
    PAGE_FRAME_ATTACHED,
    PAGE_FRAME_DETACHED,
    PAGE_FRAME_NAVIGATED,
    PAGE_FRAME_NAVIGATED_IN_DOC,
    PAGE_FRAME_STOPPED_LOADING,
    PAGE_GET_FRAME_TREE,
    PAGE_GET_LAYOUT,
    PAGE_GET_NAVIGATION_HISTORY,
    PAGE_HANDLE_DIALOG,
    PAGE_JAVASCRIPT_DIALOG_OPEN,
    PAGE_LIFECYCLE_EVENT,
    PAGE_LOAD,
    PAGE_LOAD_EVENT_FIRED,
    PAGE_NAVIGATE,
    PAGE_NAVIGATE_TO_HISTORY_ENTRY,
    PAGE_PRINT_TO_PDF,
    PAGE_RELOAD,
    PAGE_SCREENSHOT,
    PAGE_SET_BYPASS_CSP,
    PERFORMANCE_ENABLE,
    PERFORMANCE_GET_METRICS,
    PERFORMANCE_METRICS,
    RUMTIME_BINDING_CALL,
    RUNTIME_ADD_BINDING,
    RUNTIME_CALL_FUNCTION,
    RUNTIME_CONSOLE_API_CALL,
    RUNTIME_ENABLE,
    RUNTIME_EVALUATE,
    RUNTIME_EXCEPTION_THROWN,
    RUNTIME_EXECUTION_CONTEXT_CREATED,
    RUNTIME_EXECUTION_CONTEXT_DESTROYED,
    RUNTIME_EXECUTION_CONTEXTS_CLEARED,
    RUNTIME_GET_PROPERTIES,
    RUNTIME_QUERY_OBJECTS,
    RUNTIME_RELEASE_OBJECT,
    SECURITY_ENABLE,
    SECURITY_IGNORE_HTTPS_ERRORS,
    TARGET_ACTIVATE,
    TARGET_ATTACH,
    TARGET_ATTACHED,
    TARGET_CHANGED,
    TARGET_CLOSE,
    TARGET_CREATE_BROWSER_CONTEXT,
    TARGET_CREATE_TARGET,
    TARGET_CREATED,
    TARGET_DESTROYED,
    TARGET_DETACHED,
    TARGET_DISPOSE_BROWSER_CONTEXT,
    TARGET_GET_CONTEXTS,
    TARGET_INFO_CHANGED,
    TARGET_RECV_MSG,
    TARGET_SEND_DETACH,
    TARGET_SEND_MSG,
    TARGET_SET_AUTO_ATTACH,
    TARGET_SET_DISCOVER_TARGETS,
    TARGET_TARGET_CREATED,
    TARGET_TARGET_DESTROYED,
    WORKER_CREATED,
    WORKER_DESTROYED,
)
from mokr.constants.javascript import (  # noqa
    METHOD_ADD_PAGE_BINDING,
    METHOD_DELIVER_BINDING_RESULT,
    METHOD_ELEMENT_IN_VIEW,
    METHOD_EMBED_JAVASCRIPT_BY_CONTENT,
    METHOD_EMBED_JAVASCRIPT_BY_URL,
    METHOD_EMBED_STYLE_BY_CONTENT,
    METHOD_EMBED_STYLE_BY_URL,
    METHOD_EVAL_XPATH,
    METHOD_FETCH_REQUEST,
    METHOD_FOCUS_ELEMENT,
    METHOD_GET_CONTENT,
    METHOD_GET_PROPERTY,
    METHOD_SCROLL_INTO_VIEW,
    METHOD_SELECT_VALUES,
    METHOD_SET_CONTENT,
    METHOD_WAIT_FOR_PREDICATE_PAGE,
    METHOD_WAIT_FOR_XPATH_OR_SELECTOR,
)
from mokr.constants.typing import (  # noqa
    HTTP_CACHE_TYPES,
    HTTP_METHODS,
    LIFECYCLE_EVENTS,
    REFERRER_POLICIES,
)
from mokr.constants.http import (  # noqa
    HTTP_RESPONSE_CODES_TEXT_MAP,
    NETWORK_ERROR_CODES_TO_REASONS,
)
