CLOSE = "close"
DISCONNECTED = "disconnected"
ERROR = "error"
LIFECYCLE_EVENT = "lifecycleevent"
METRICS = "metrics"

BROWSER_CLOSE = "Browser.close"
BROWSER_GET_VERSION = "Browser.getVersion"

DOM_BOX_MODEL = "DOM.getBoxModel"
DOM_CONTENT_QUADS = "DOM.getContentQuads"
DOM_DESCRIBE_NODE = "DOM.describeNode"
DOM_INPUT_FILES = "DOM.setFileInputFiles"
DOM_LOADED = "domcontentloaded"

EMULATION_DISABLE_SCRIPT_EXECUTION = "Emulation.setScriptExecutionDisabled"
EMULATION_ENABLE_TOUCH = "Emulation.setTouchEmulationEnabled"
EMULATION_OVERRIDE_BACKGROUND = "Emulation.setDefaultBackgroundColorOverride"
EMULATION_OVERRIDE_METRICS = "Emulation.setDeviceMetricsOverride"
EMULATION_SET_EMULATED_MEDIA = "Emulation.setEmulatedMedia"

FETCH_AUTH_REQD = "Fetch.authRequired"
FETCH_CONTINUE = "Fetch.continueRequest"
FETCH_CONTINUE_AUTH = "Fetch.continueWithAuth"
FETCH_DISABLED = "Fetch.disable"
FETCH_DOMAIN_REQUEST_SENT = "FetchDomain.requestSent"
FETCH_ENABLED = "Fetch.enable"
FETCH_FAIL = "Fetch.failRequest"
FETCH_FULFILL = "Fetch.fulfillRequest"
FETCH_REQUEST_PAUSED = "Fetch.requestPaused"

FRAME_ATTACHED = "frameattached"
FRAME_DETACHED = "framedetached"
FRAME_NAVIGATED = "framenavigated"
FRAME_NAVIGATED_IN_DOC = "framenavigatedwithindocument"

INPUT_INSERT_TEXT = "Input.insertText"
INPUT_KEY = "Input.dispatchKeyEvent"
INPUT_MOUSE = "Input.dispatchMouseEvent"
INPUT_TOUCH = "Input.dispatchTouchEvent"

INSPECTOR_TARGET_CRASHED = "Inspector.targetCrashed"

LOG_ENABLE = "Log.enable"
LOG_ENTRY_ADDED = "Log.entryAdded"

NETWORK_CACHE_DISABLE = "Network.setCacheDisabled"
NETWORK_DELETE_COOKIES = "Network.deleteCookies"
NETWORK_EMULATE_NETWORK_CONDITIONS = "Network.emulateNetworkConditions"
NETWORK_ENABLE = "Network.enable"
NETWORK_EXTRA_HEADERS = "Network.setExtraHTTPHeaders"
NETWORK_GET_ALL_COOKIES = "Network.getAllCookies"
NETWORK_GET_COOKIES = "Network.getCookies"
NETWORK_GET_RESPONSE_BODY = "Network.getResponseBody"
NETWORK_LOADING_FAILED = "Network.loadingFailed"
NETWORK_LOADING_FINISHED = "Network.loadingFinished"
NETWORK_MGR_REQUEST = "NetworkManager.Request"
NETWORK_MGR_REQUEST_FAILED = "NetworkManager.RequestFailed"
NETWORK_MGR_REQUEST_FINISHED = "NetworkManager.RequestFinished"
NETWORK_MGR_REQUEST_FROM_CACHE = "NetworkManager.RequestServedFromCache"
NETWORK_MGR_RESPONSE = "NetworkManager.Response"
NETWORK_REQUEST = "request"
NETWORK_REQUEST_FAILED = "requestfailed"
NETWORK_REQUEST_FINISHED = "requestfinished"
NETWORK_REQUEST_SERVED_FROM_CACHE = "Network.requestServedFromCache"
NETWORK_REQUEST_WILL_BE_SENT = "Network.requestWillBeSent"
NETWORK_RESPONSE = "response"
NETWORK_RESPONSE_RECVD = "Network.responseReceived"
NETWORK_RESPONSE_RECVD_EXTRA = "Network.responseReceivedExtraInfo"
NETWORK_SET_COOKIES = "Network.setCookies"
NETWORK_USER_AGENT_OVERRIDE = "Network.setUserAgentOverride"

PAGE_ADD_SCRIPT_TO_EVAL = "Page.addScriptToEvaluateOnNewDocument"
PAGE_BRING_TO_FRONT = "Page.bringToFront"
PAGE_CLOSE = "Page.close"
PAGE_CONSOLE = "console"
PAGE_DIALOG = "dialog"
PAGE_DOM_LOADED = "Page.domContentEventFired"
PAGE_ENABLE = "Page.enable"
PAGE_ENABLE_LIFECYCLE_EVENTS = "Page.setLifecycleEventsEnabled"
PAGE_ERROR = "pageerror"
PAGE_FRAME_ATTACHED = "Page.frameAttached"
PAGE_FRAME_DETACHED = "Page.frameDetached"
PAGE_FRAME_NAVIGATED = "Page.frameNavigated"
PAGE_FRAME_NAVIGATED_IN_DOC = "Page.navigatedWithinDocument"
PAGE_FRAME_STOPPED_LOADING = "Page.frameStoppedLoading"
PAGE_GET_FRAME_TREE = "Page.getFrameTree"
PAGE_GET_LAYOUT = "Page.getLayoutMetrics"
PAGE_GET_NAVIGATION_HISTORY = "Page.getNavigationHistory"
PAGE_HANDLE_DIALOG = "Page.handleJavaScriptDialog"
PAGE_JAVASCRIPT_DIALOG_OPEN = "Page.javascriptDialogOpening"
PAGE_LIFECYCLE_EVENT = "Page.lifecycleEvent"
PAGE_LOAD = "load"
PAGE_LOAD_EVENT_FIRED = "Page.loadEventFired"
PAGE_NAVIGATE = "Page.navigate"
PAGE_NAVIGATE_TO_HISTORY_ENTRY = "Page.navigateToHistoryEntry"
PAGE_PRINT_TO_PDF = "Page.printToPDF"
PAGE_RELOAD = "Page.reload"
PAGE_SCREENSHOT = "Page.captureScreenshot"
PAGE_SET_BYPASS_CSP = "Page.setBypassCSP"

PERFORMANCE_ENABLE = "Performance.enable"
PERFORMANCE_GET_METRICS = "Performance.getMetrics"
PERFORMANCE_METRICS = "Performance.metrics"

RUMTIME_BINDING_CALL = "Runtime.bindingCalled"
RUNTIME_ADD_BINDING = "Runtime.addBinding"
RUNTIME_CALL_FUNCTION = "Runtime.callFunctionOn"
RUNTIME_CONSOLE_API_CALL = "Runtime.consoleAPICalled"
RUNTIME_ENABLE = "Runtime.enable"
RUNTIME_EVALUATE = "Runtime.evaluate"
RUNTIME_EXCEPTION_THROWN = "Runtime.exceptionThrown"
RUNTIME_EXECUTION_CONTEXTS_CLEARED = "Runtime.executionContextsCleared"
RUNTIME_EXECUTION_CONTEXT_CREATED = "Runtime.executionContextCreated"
RUNTIME_EXECUTION_CONTEXT_DESTROYED = "Runtime.executionContextDestroyed"
RUNTIME_GET_PROPERTIES = "Runtime.getProperties"
RUNTIME_QUERY_OBJECTS = "Runtime.queryObjects"
RUNTIME_RELEASE_OBJECT = "Runtime.releaseObject"

SECURITY_ENABLE = "Security.enable"
SECURITY_IGNORE_HTTPS_ERRORS = "Security.setIgnoreCertificateErrors"

TARGET_ACTIVATE = "Target.activateTarget"
TARGET_ATTACH = "Target.attachToTarget"
TARGET_ATTACHED = "Target.attachedToTarget"
TARGET_CHANGED = "targetchanged"
TARGET_CLOSE = "Target.closeTarget"
TARGET_CREATED = "targetcreated"
TARGET_CREATE_BROWSER_CONTEXT = "Target.createBrowserContext"
TARGET_CREATE_TARGET = "Target.createTarget"
TARGET_DESTROYED = "targetdestroyed"
TARGET_DETACHED = "Target.detachedFromTarget"
TARGET_DISPOSE_BROWSER_CONTEXT = "Target.disposeBrowserContext"
TARGET_GET_CONTEXTS = "Target.getBrowserContexts"
TARGET_INFO_CHANGED = "Target.targetInfoChanged"
TARGET_RECV_MSG = "Target.receivedMessageFromTarget"
TARGET_SEND_DETACH = "Target.detachFromTarget"
TARGET_SEND_MSG = "Target.sendMessageToTarget"
TARGET_SET_AUTO_ATTACH = "Target.setAutoAttach"
TARGET_SET_DISCOVER_TARGETS = "Target.setDiscoverTargets"
TARGET_TARGET_CREATED = "Target.targetCreated"
TARGET_TARGET_DESTROYED = "Target.targetDestroyed"

WORKER_CREATED = "workercreated"
WORKER_DESTROYED = "workerdestroyed"
